#
# Simulation exercise
#
remove(list = ls())
graphics.off()
parnames <- c('p', 'q', 'm', 'b1', 'b2', 'b3')
datos <- read.csv('C:/Users/SantiagoM/Google Drive/Estudio/Bass/Datos/Simulation.csv')
Time <- nrow(datos)
X <- rbind(0, apply(as.matrix(datos[, c('x1', 'x2', 'x3')]), 2, function(x) log(x/x[1])))
M <- 50000

# Log-likelihood function with explicit restrictions
ll_gbm <- function(pars, y, M, X) {
    # a <- pars[1]
    # b <- pars[2]
    # cc <- pars[3]
    p <- pars[1]
    q <- pars[2]
    m <- pars[3]
    betas <- pars[4:length(pars)]
    Time <- length(y)
    tt <- 0:Time
    x <- c(y, M - sum(y))
    Xt <- c(tt + X%*%betas)
    # Ft <- cc*(1 - exp(-b*Xt))/(1 + a*exp(-b*Xt))
    Ft <- (m/M)*(1 - exp(-(p+q)*Xt))/(1 + (q/p)*exp(-(p+q)*Xt))
    # if (a < 0 | b < 0 | cc < 0 | cc > 1) {
    if (p < 0 | p > 1 | q < 0 | q > 1 | m < 0) {
        res <- Inf
    } else {
        res <- -x[Time + 1]*log(1 - Ft[Time + 1]) - sum(x[tt[-1]]*log(diff(Ft)))
    }
    return(res)
}

# Function for fitted values with covariates
fittedbayeswc <- function(pars, X) {
    p <- pars[1]; q <- pars[2]; m <- pars[3]; betas <- pars[4:6]
    Time <- nrow(X); unit <- 1:Time
    xdata <- c(unit + X%*%betas)
    Ft <- c(0, ((1 - exp(-xdata*(p+q)))/(1 + (q/p)*exp(-xdata*(p+q)))))
    res <- m*diff(Ft)
    return(res)
}

# Sales measures
sales_measures <- function(fs) {
    peak <- which.max(fs)
    take1 <- mean(bede(x = 1:peak, y = fs[1:peak], index = 0)$iters[, 'EDE'])
    take2 <- mean(bese(x = 1:peak, y = fs[1:peak], index = 0)$iters[, 'ESE'])
    take3 <- ede(x = 1:peak, y = fs[1:peak], index = 0)[, 'chi']
    take4 <- ese(x = 1:peak, y = fs[1:peak], index = 0)[, 'chi']

    sat1 <- mean(bede(x = peak:Time, y = fs[peak:Time], index = 1)$iters[, 'EDE'])
    sat2 <- mean(bese(x = peak:Time, y = fs[peak:Time], index = 1)$iters[, 'ESE'])
    sat3 <- ede(x = peak:Time, y = fs[peak:Time], index = 1)[, 'chi']
    sat4 <- ese(x = peak:Time, y = fs[peak:Time], index = 1)[, 'chi']

    return(list(peak = peak, takeoff = c(take1, take2, take3, take4),
                saturation = c(sat1, sat2, sat3, sat4)))
}

# Data amounts and results list
Tvec <- c(5, 8, 9, 10, 15, 20, 25)
reslist <- rep(list(NA), length(Tvec))
names(reslist) <- Tvec

coefmat <- matrix(NA, nrow = length(Tvec), ncol = length(parnames))
semat <- matrix(NA, nrow = length(Tvec), ncol = length(parnames))
fittedmat <- matrix(NA, nrow = length(Tvec), ncol = Time)
colnames(coefmat) <- colnames(semat) <- parnames
rownames(coefmat) <- rownames(semat) <- rownames(fittedmat) <- Tvec

for (i in 1:length(Tvec)) {
    Time <- Tvec[i]
    unit <- 1:Time
    datosT <- datos[unit, ]
    XT <- X[1:(Time + 1), ]

    # OLS regression for starting values
    ols <- lm(Ss ~ Nt + Nt2, data = datosT) # Three data points necessary
    a <- coef(ols)[3]; b <- coef(ols)[2]; c <- coef(ols)[1]
    a <- unname(a); b <- unname(b); c <- unname(c)
    mest <- (-b - sqrt(b^2 - 4*a*c))/(2*a)
    pest <- c/mest; qest <- -a*mest
    parswc <- list(p = pest, q = qest, m = mest, b = rep(0, ncol(XT)))
    pars.vwc <- unlist(parswc)

    # Estimation with covariates
    print(Time)
    fn_gbm <- optim(par = pars.vwc, fn = ll_gbm, hessian = TRUE, y = datosT$Ss, M = M, X = XT)

    # Saving results
    coefmat[i, ] <- fn_gbm$par
    semat[i, ] <- sqrt(diag(solve(fn_gbm$hessian)))
    fittedmat[i, ] <- fittedbayeswc(coefmat[i, ], X[-1, , drop = FALSE])
}

# Forecast matrices
fittedmat <- t(fittedmat)
cumfittedmat <- apply(fittedmat, 2, cumsum)

# Summary measures
t(coefmat)
t(semat)
apply(fittedmat, 2, sales_measures)

# Cumulative forecast plots
reppoints <- c(0, 5, 10, 25)
setEPS()
postscript('C:/Users/SantiagoM/Google Drive/Estudio/Bass/Tex/FEx1.eps')
plot(datos$Yt, type = 'l', main = 'Cumulative Sales', xlab = 'Years', ylab = 'Sales',
     ylim = c(0, max(fittedmat)))
sapply(1:length(Tvec), function(n) lines(fittedmat[ , n], col = 1 + n, lty = 2))
legend(x = 'topleft', legend = c('Simulated', paste(reppoints, 'data points')),
       col = c(1, reppoints + 2), lty = c(1, rep(2, length(reppoints))))
dev.off()

# #
# # Adaptative plots
# #
#
# # Data matrices and plot limits
# zmatx <- sapply(1:6, function(i) density(reslist$`0`$Chains[ , i])$x)
# zmaty <- sapply(1:6, function(i) density(reslist$`0`$Chains[ , i])$y)
# zlimx <- rbind(apply(zmatx, 2, min), apply(zmatx, 2, max))
# zlimy <- rbind(apply(zmaty, 2, min), apply(zmaty, 2, max))
#
# densize <- length(density(reslist[[2]]$Chains[ , 1])$x)
# tmatx <- matrix(NA, nrow = densize*(length(Tvec) - 1), ncol = 6)
# tmaty <- matrix(NA, nrow = densize*(length(Tvec) - 1), ncol = 6)
# tlimx <- matrix(NA, nrow = 2*(length(Tvec) - 1), ncol = 6)
# tlimy <- matrix(NA, nrow = 2*(length(Tvec) - 1), ncol = 6)
#
# for (j in 1:6) {
#     for (i in 1:(length(Tvec) - 1)) {
#         tmatx[(densize*(i - 1) + 1):(densize*i), j] <- density(reslist[[i + 1]]$Chains[ , j])$x
#         tmaty[(densize*(i - 1) + 1):(densize*i), j] <- density(reslist[[i + 1]]$Chains[ , j])$y
#         tlimx[(2*i - 1):(2*i), j] <- c(min(tmatx[(densize*(i - 1) + 1):(densize*i), j]), max(tmatx[(densize*(i - 1) + 1):(densize*i), j]))
#         tlimy[(2*i - 1):(2*i), j] <- c(min(tmaty[(densize*(i - 1) + 1):(densize*i), j]), max(tmaty[(densize*(i - 1) + 1):(densize*i), j]))
#     }
# }
#
# # Plots
# for (j in 1:3) {
#     dev.new()
#     plot(zmatx[ , j], zmaty[ , j], type = 'l', main = paste('Prior and posterior distributions for', parnames[j]),
#          xlab = 'Parameter Space', ylab = 'Density', xlim = c(min(c(zlimx[ , j], tlimx[ , j])), max(c(zlimx[ , j], tlimx[ , j])))
#          # ylim = c(min(c(zlimy[ , j], tlimy[ , j])), max(c(zlimy[ , j], tlimy[ , j])))
#          )
#     for (i in 1:(length(Tvec) - 1)) {
#         lines(tmatx[(densize*(i - 1) + 1):(densize*i), j], tmaty[(densize*(i - 1) + 1):(densize*i), j], col = i + 1)
#     }
#     abline(v = pobpars[j], lty = 2)
# #     legend(x = 'topleft', legend = c('Prior', paste(Tvec[-1], 'data points')),
# #            col = c(1, 2:length(Tvec)), lty = 1)
# }
#
# for (j in 4:6) {
#     dev.new()
#     plot(zmatx[ , j], zmaty[ , j], type = 'l', main = substitute(paste(x, beta[n]), env = list(x = 'Prior and posterior distributions for ', n = j - 3)),
#          xlab = 'Parameter Space', ylab = 'Density', xlim = c(min(c(zlimx[ , j], tlimx[ , j])), max(c(zlimx[ , j], tlimx[ , j])))
#          # ylim = c(min(c(zlimy[ , j], tlimy[ , j])), max(c(zlimy[ , j], tlimy[ , j])))
#          )
#     for (i in 1:(length(Tvec) - 1)) {
#         lines(tmatx[(densize*(i - 1) + 1):(densize*i), j], tmaty[(densize*(i - 1) + 1):(densize*i), j], col = i + 1)
#     }
#     abline(v = pobpars[j], lty = 2)
# #     legend(x = 'topleft', legend = c('Prior', paste(Tvec[-1], 'data points')),
# #            col = c(1, 2:length(Tvec)), lty = 1)
# }

#
# Mean Errors
#

msefit <- sapply(1:length(Tvec), function(n) mean((c(fittedmat[1, n], diff(fittedmat[ , n])) - datos$Ss)^2))
maefit <- sapply(1:length(Tvec), function(n) mean(abs(c(fittedmat[1, n], diff(fittedmat[ , n])) - datos$Ss)))
msecumfit <- sapply(1:length(Tvec), function(n) mean((fittedmat[ , n] - datos$Yt)^2))
maecumfit <- sapply(1:length(Tvec), function(n) mean(abs(fittedmat[ , n] - datos$Yt)))

# One step-ahead forecast
osaf <- diag(fittedmat[,-length(Tvec)])
(osaf - datos$Yt)^2
